# Snafu: Snake Functions - C Executor

import subprocess
import os
import time

def execute(func, funcargs, envvars, sourceinfos):
	#print(">> call", func, funcargs, envvars, sourceinfos.source)

	ccmd = "snafulib/executors/c/cexec {} {}".format(sourceinfos.source, " ".join(funcargs))
	stime = time.time()
	out, err = subprocess.Popen(ccmd, shell=True, stdout=subprocess.PIPE).communicate()
	dtime = (time.time() - stime) * 1000

	try:
		out = out.decode("utf-8").split("\n")[-2]
		success = True
	except:
		out = ""
		success = False

	return dtime, success, out



# Snafu: Snake Functions - JavaScript Executor

import subprocess
import time
import os.path

def execute(func, funcargs, envvars, sourceinfos):
	sourcemodule = sourceinfos.source[:-3]

	if "." in func:
		func = func.split(".")[1]

	# exports hack
	sourcemodulemod = sourcemodule + ".js.export"
	if not os.path.isfile(sourcemodulemod):
		code = open(sourcemodule + ".js").read()
		if not "exports" in code:
			f = open(sourcemodulemod, "w")
			f.write(code)
			f.write("\n")
			f.write("function mainwrapper(input){console.log(main(input));}\n")
			f.write("exports.main = mainwrapper;\n")
			sourcemodule = sourcemodulemod
	else:
		sourcemodule = sourcemodulemod

	# message hack
	if not "{" in funcargs[0]:
		funcargs[0] = "\"{\"body\": {\"message\": \"%s\"}}\"" % funcargs[0]

	if sourcemodulemod != sourcemodule:
		cmd = "nodejs -e 'require(\"./%s\").%s(%s, {\"status\": function(x){return {\"send\": function(x){console.log(\"RET:\" + x)}}}})'" % (sourcemodule, func, funcargs[0])
	else:
		cmd = "nodejs -e 'require(\"./%s\").%s(%s)'" % (sourcemodule, func, funcargs[0])

	stime = time.time()
	p = subprocess.run(cmd, stdout=subprocess.PIPE, shell=True)
	out = p.stdout.decode("utf-8")
	dtime = time.time() - stime

	success = False
	res = []
	for line in out.split("\n"):
		if sourcemodulemod != sourcemodule:
			if line.startswith("RET:"):
				success = True
				res = line[4:]
				break
		else:
			success = True
			res = line
			break

	return dtime, success, res



# Snafu: Snake Functions - Java Executor

#import json
#import pickle
#import base64
import subprocess
import os
import time

def execute(func, funcargs, envvars, sourceinfos):
	classname = os.path.basename(sourceinfos.source).split(".")[0]
	methodname = func.split(".")[1]

	#print(">> call", classname, methodname, func, funcargs, envvars, sourceinfos.source)

	#classname, methodname = func.split(".")
	javacmd = "java -cp snafulib/executors/java/:{} JavaExec {} {} {}".format(os.path.dirname(sourceinfos.source), classname, methodname, " ".join(funcargs))
	#print(">> javacmd", javacmd)
	stime = time.time()
	out, err = subprocess.Popen(javacmd, shell=True, stdout=subprocess.PIPE).communicate()
	dtime = (time.time() - stime) * 1000
	#print(">> result", out)

	out = out.decode("utf-8").split("\n")[-2]

	return dtime, True, out



# Snafu: Snake Functions - Python2 Stateful Executor

#import json
import subprocess
#import pickle
#import base64
import threading
import queue
import os
import time

firstcall = {}

def asynccommand(command, commandprocess, qerr, qout, inputline, wait=True):
	def queuethread(out, queue):
		try:
			for line in iter(out.readline, b""):
				line = line.decode("utf-8").strip()
				queue.put(line)
		except Exception as e:
			pass

	if not commandprocess:
		commandprocess = subprocess.Popen(command, shell=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE, stdin=subprocess.PIPE)
	commandprocess.stdin.write(bytes(inputline + "\n", "utf-8"))
	commandprocess.stdin.flush()

	if not qout and not qerr:
		qout = queue.Queue()
		qerr = queue.Queue()
		tout = threading.Thread(target=queuethread, args=(commandprocess.stdout, qout))
		tout.daemon = True
		tout.start()
		terr = threading.Thread(target=queuethread, args=(commandprocess.stderr, qerr))
		terr.daemon = True
		terr.start()

	if not wait:
		return commandprocess, qerr, qout, ""

	results = ""
	had_output = False
	counter = 0
	while True:
		both_empty = True
		try:
			line = qout.get(timeout=0.001)
		except queue.Empty:
			pass
		else:
			had_output = True
			both_empty = False
			results += line
		try:
			line = qerr.get(timeout=0.001)
		except queue.Empty:
			pass
		else:
			#had_output = True
			both_empty = False
			#results += "(err)" + line
		if had_output and both_empty:
			return commandprocess, qerr, qout, results
		#counter += 1
		#if counter > 100:
		#	return commandprocess, qerr, qout, results

def execute_internal(func, funcargs, envvars, sourceinfos):
	global firstcall

	import_stmt1 = "import os"
	import_stmt2 = "os.chdir('{}')".format(os.path.dirname(sourceinfos.source))
	module = os.path.basename(sourceinfos.source)[:-3]
	import_stmt3 = "import {}".format(module)

	funcargsfmt = []
	for funcarg in funcargs:
		if type(funcarg) in (dict, list, int, float):
			funcargsfmt.append(str(funcarg))
		else:
			funcargsfmt.append("\"{}\"".format(str(funcarg)))
	funcargsfmt = ",".join(funcargsfmt)
	exec_stmt = "{}.{}({})".format(module, func.__name__, funcargsfmt)

	callid = threading.get_ident()
	if not callid in firstcall:
		proc, qerr, qout, r1 = asynccommand("python2 -iu", None, None, None, import_stmt1, wait=False)
		proc, qerr, qout, r2 = asynccommand(None, proc, qerr, qout, import_stmt2, wait=False)
		proc, qerr, qout, r3 = asynccommand(None, proc, qerr, qout, import_stmt3, wait=False)
		firstcall[callid] = proc, qerr, qout
	else:
		proc, qerr, qout = firstcall[callid]
	proc, qerr, qout, r4 = asynccommand(None, proc, qerr, qout, exec_stmt)

	#for i, funcarg in enumerate(funcargs):
	#	if "__class__" in dir(funcarg) and funcarg.__class__.__name__ == "SnafuContext":
	#		funcargs[i] = "pickle:" + base64.b64encode(pickle.dumps(funcarg, protocol=2)).decode("utf-8")
	# + envvars
	#funcargs = json.dumps(funcargs)
	#envvars = json.dumps(envvars)

	return r4.replace("'", "\"")

def execute(func, funcargs, envvars, sourceinfos):
	stime = time.time()

	try:
		res = execute_internal(func, funcargs, envvars, sourceinfos)
		success = True
	except Exception as e:
		res = e
		success = False
	dtime = (time.time() - stime) * 1000

	return dtime, success, res



# Snafu: Snake Functions - Web Connector

import flask
import threading
import os
import configparser

app = flask.Flask("snafu")

gcb = None

@app.route("/invoke/<function>")
def invoke(function):
	response = gcb(function)
	if not response:
		flask.abort(500)
	return response

def initinternal(function, configpath):
	connectconfig = None
	if not configpath:
		configpath = "snafu.ini"
	if not function:
		function = "snafu"
	if os.path.isfile(configpath):
		config = configparser.ConfigParser()
		config.read(configpath)
		if function in config and "connector.web" in config[function]:
			connectconfig = int(config[function]["connector.web"])

	if connectconfig:
		app.run(host="0.0.0.0", port=connectconfig)

def init(cb, function=None, configpath=None):
	global gcb
	gcb = cb

	t = threading.Thread(target=initinternal, daemon=True, args=(function, configpath))
	t.start()



# Snafu: Snake Functions - Cron Connector

import threading
import time
import os
import configparser

gcb = None

def initinternal(function, configpath):
	connectconfig = None
	if not configpath:
		configpath = "snafu.ini"
	if not function:
		function = "snafu"
	if os.path.isfile(configpath):
		config = configparser.ConfigParser()
		config.read(configpath)
		if function in config and "connector.cron" in config[function]:
			connectconfig = int(config[function]["connector.cron"])

	if connectconfig:
		while True:
			time.sleep(connectconfig)
			response = gcb(function, event="{}")

def init(cb, function=None, configpath=None):
	global gcb
	gcb = cb

	t = threading.Thread(target=initinternal, daemon=True, args=(function, configpath))
	t.start()



# Snafu: Snake Functions - SQLite Logger

import sqlite3
import os
import configparser

def log(source, function, duration, success):
	logurl = "/tmp/snafu.sqlite"
	if os.path.isfile("snafu.ini"):
		config = configparser.ConfigParser()
		config.read("snafu.ini")
		if "snafu" in config and "logger.sqlite" in config["snafu"]:
			logurl = config["snafu"]["logger.sqlite"]

	init = False
	if not os.path.isfile(logurl):
		init = True
	conn = sqlite3.connect(logurl)
	c = conn.cursor()
	if init:
		c.execute("CREATE TABLE log (source text, function text, duration real, success bool)")
	c.execute("INSERT INTO log (source, function, duration, success) VALUES ('{}', '{}', {}, '{}')".format(source, function, duration, success))
	conn.commit()
	conn.close()



# python3 setup.py sdist
# python3 setup.py bdist_wheel
# python3 setup.py bdist_egg
# twine upload dist/*.* [-r testpypi]
# -> automated in tools/make-dist.sh

from setuptools import setup
import os

setup(
	name="snafu",
	description="Swiss Army Knife of Serverless Computing",
	version="0.0.1",
	url="https://github.com/serviceprototypinglab/snafu",
	author="Josef Spillner",
	author_email="josef.spillner@zhaw.ch",
	license="Apache 2.0",
	classifiers=[
		"Development Status :: 2 - Pre-Alpha",
		"Environment :: Console",
		"Environment :: No Input/Output (Daemon)",
		"Intended Audience :: Science/Research",
		"License :: OSI Approved :: Apache Software License",
		"Programming Language :: Python :: 3",
		"Topic :: Internet :: WWW/HTTP :: WSGI :: Middleware"
	],
	keywords="cloud faas serverless functions",
	packages=["snafulib", "snafulib.executors", "snafulib.loggers", "snafulib.authenticators", "snafulib.connectors"],
	scripts=["snafu", "snafu-import", "snafu-accounts", "snafu-control"],
	install_requires=["flask"]
)



# Snafu: Snake Functions - OpenShift Executor

import requests
import os
import configparser
import subprocess

container = "jszhaw/snafu"

endpoints = {}

def executecontrol(flaskrequest, tenant):
	if not tenant in endpoints:
		username = os.getenv("OPENSHIFT_USERNAME")
		password = os.getenv("OPENSHIFT_PASSWORD")
		password = os.getenv("OPENSHIFT_PROJECT")
		if not username or not password or not project:
			return
		os.system("oc login https://console.appuio.ch/ --username={} --password={}".format(username, password))
		os.system("oc project {}".format(project))
		os.system("oc new-app --name snafu-{} jszhaw/snafu".format(tenant))
		p = subprocess.run("oc status | grep svc/snafu-{} | cut -d " " -f 3".format(tenant), shell=True, stdout=subprocess.PIPE)

		endpoints[tenant] = "http://{}".format(p.decode("utf-8"))
		# FIXME: mounting the tenant's volume container to /opt/functions-local
	endpoint = endpoints[tenant]

	headers = {}
	headers["X-Amz-Date"] = flaskrequest.headers.get("X-Amz-Date")

	data = flaskrequest.data.decode("utf-8")
	#method=r.method -> requests.post

	reply = requests.post(endpoint + flaskrequest.path, data=data, headers=headers)

	if reply.status_code == 200:
		return reply.content.decode("utf-8")
	else:
		return
